/*
    _____               __  ___          __            ____        _      __
   / ___/__ ___ _  ___ /  |/  /__  ___  / /_____ __ __/ __/_______(_)__  / /_
  / (_ / _ `/  ' \/ -_) /|_/ / _ \/ _ \/  '_/ -_) // /\ \/ __/ __/ / _ \/ __/
  \___/\_,_/_/_/_/\__/_/  /_/\___/_//_/_/\_\\__/\_, /___/\__/_/ /_/ .__/\__/
                                               /___/             /_/
                                             
  See Copyright Notice in gmMachine.h

*/

#include "stdafx.h"
#include "gmDebugger.h"


gmDebuggerSession::gmDebuggerSession()
{
  m_outSize = 256;
  m_out = malloc(256);
  m_outCursor = 0;
  m_in = NULL;
  m_inCursor = m_inSize = NULL;
}


gmDebuggerSession::~gmDebuggerSession()
{
  if(m_out) free(m_out);
}


void gmDebuggerSession::Update()
{
  for(;;)
  {
    m_in = m_pumpMessage(this, m_inSize);
    if(m_in == NULL) break;
    m_inCursor = 0;

    int id, pa, pb, pc;
    const char * sa, * sb, * sc;

    Unpack(id);
    switch(id)
    {
      case 'dbrk' :
        Unpack(id).Unpack(pa).Unpack(pb);
        gmDebuggerBreak(this, id, pa, pb);
        break;
      case 'drun' :
        Unpack(id);
        gmDebuggerRun(this, id);
        break;
      case 'dstp' :
        Unpack(id);
        gmDebuggerStop(this, id);
        break;
      case 'dsrc' :
        Unpack(id).Unpack(sa).Unpack(sb);
        gmDebuggerSource(this, id, sa, sb);
        break;
      case 'dexc' :
        Unpack(id);
        gmDebuggerException(this, id);
        break;
      case 'dctx' :
        Unpack(id).Unpack(pa); // thread id, callframe
        gmDebuggerBeginContext(this, id, pa);
        for(;;)
        {
          Unpack(id);
          if(id == 'call')
          {
            Unpack(id).Unpack(sa).Unpack(pa).Unpack(pb).Unpack(sb).Unpack(sc).Unpack(pc);
            gmDebuggerContextCallFrame(this, id, sa, pa, pb, sb, sc, pc);
          }
          else if(id == 'vari')
          {
            Unpack(sa).Unpack(sb).Unpack(pa);
            gmDebuggerContextVariable(this, sa, sb, pa);
          }
          else if(id == 'done') break;
          else break;
        }
        gmDebuggerEndContext(this);
        break;
      case 'dsri' :
        // todo
        break;
      case 'dthi' :
        gmDebuggerBeginThreadInfo(this);
        for(;;)
        {
          Unpack(id);
          if(id == 'thri')
          {
            Unpack(pa).Unpack(pb);
            gmDebuggerThreadInfo(this, pa, pb);
          }
          else if(id == 'done') break;
          else break;
        }
        gmDebuggerEndThreadInfo(this);
        break;
      case 'derr' :
        Unpack(sa);
        gmDebuggerError(this, sa);
        break;
      case 'dmsg' :
        Unpack(sa);
        gmDebuggerMessage(this, sa);
        break;
      case 'dack' :
        Unpack(pa).Unpack(pb);
        gmDebuggerAck(this, pa, pb);
        break;
      case 'dend' :
        gmDebuggerQuit(this);
        break;
      default:;
    }
  }
}


bool gmDebuggerSession::Open()
{
  m_outCursor = 0;
  return true;
}


bool gmDebuggerSession::Close()
{
  return true;
}


gmDebuggerSession &gmDebuggerSession::Pack(int a_val)
{
  Need(4);
  memcpy((char *) m_out + m_outCursor, &a_val, 4);
  m_outCursor += 4;
  return *this;
}


gmDebuggerSession &gmDebuggerSession::Pack(const char * a_val)
{
  if(a_val)
  {
    int len = strlen(a_val) + 1;
    Need(len);
    memcpy((char *) m_out + m_outCursor, a_val, len);
    m_outCursor += len;
  }
  else
  {
    Need(1);
    memcpy((char *) m_out + m_outCursor, "", 1);
    m_outCursor += 1;
  }
  return *this;
}


void gmDebuggerSession::Send()
{
  m_sendMessage(this, m_out, m_outCursor);
  m_outCursor = 0;
}


gmDebuggerSession &gmDebuggerSession::Unpack(int &a_val)
{
  if(m_inCursor + 4 <= m_inSize)
  {
    memcpy(&a_val, (const char *) m_in + m_inCursor, 4);
    m_inCursor += 4;
  }
  else
  {
    a_val = 0;
  }
  return *this;
}


gmDebuggerSession &gmDebuggerSession::Unpack(const char * &a_val)
{
  a_val = (const char *) m_in + m_inCursor;
  m_inCursor += strlen(a_val) + 1;
  return *this;
}


void gmDebuggerSession::Need(int a_bytes)
{
  if((m_outCursor + a_bytes) >= m_outSize)
  {
    int newSize = m_outSize + a_bytes + 256;
    void * buffer = malloc(newSize);
    memcpy(buffer, m_out, m_outCursor);
    free(m_out);
    m_out = buffer;
    m_outSize = newSize;
  }
}


void gmMachineRun(gmDebuggerSession * a_session, int a_threadId)
{
  a_session->Pack('mrun').Pack(a_threadId).Send();
}

void gmMachineStepInto(gmDebuggerSession * a_session, int a_threadId)
{
  a_session->Pack('msin').Pack(a_threadId).Send();
}

void gmMachineStepOver(gmDebuggerSession * a_session, int a_threadId)
{
  a_session->Pack('msov').Pack(a_threadId).Send();
}

void gmMachineStepOut(gmDebuggerSession * a_session, int a_threadId)
{
  a_session->Pack('msou').Pack(a_threadId).Send();
}

void gmMachineGetContext(gmDebuggerSession * a_session, int a_threadId, int a_callframe)
{
  a_session->Pack('mgct').Pack(a_threadId).Pack(a_callframe).Send();
}

void gmMachineGetSource(gmDebuggerSession * a_session, int a_sourceId)
{
  a_session->Pack('mgsr').Pack(a_sourceId).Send();
}

void gmMachineGetSourceInfo(gmDebuggerSession * a_session)
{
  a_session->Pack('mgsi').Send();
}

void gmMachineGetThreadInfo(gmDebuggerSession * a_session)
{
  a_session->Pack('mgti').Send();
}

void gmMachineGetVariableInfo(gmDebuggerSession * a_session, int a_variableId)
{
  a_session->Pack('mgvi').Pack(a_variableId).Send();
}

void gmMachineSetBreakPoint(gmDebuggerSession * a_session, int a_responseId, int a_sourceId, int a_lineNumber, int a_threadId, int a_enabled)
{
  a_session->Pack('msbp').Pack(a_responseId).Pack(a_sourceId).Pack(a_lineNumber).Pack(a_threadId).Pack(a_enabled).Send();
}

void gmMachineBreak(gmDebuggerSession * a_session, int a_threadId)
{
  a_session->Pack('mbrk').Pack(a_threadId).Send();
}

void gmMachineQuit(gmDebuggerSession * a_session)
{
  a_session->Pack('mend').Send();
  Sleep(500);
}