/*
    _____               __  ___          __            ____        _      __
   / ___/__ ___ _  ___ /  |/  /__  ___  / /_____ __ __/ __/_______(_)__  / /_
  / (_ / _ `/  ' \/ -_) /|_/ / _ \/ _ \/  '_/ -_) // /\ \/ __/ __/ / _ \/ __/
  \___/\_,_/_/_/_/\__/_/  /_/\___/_//_/_/\_\\__/\_, /___/\__/_/ /_/ .__/\__/
                                               /___/             /_/
                                             
  See Copyright Notice in gmMachine.h

*/

#ifndef _GMSTREAM_H_
#define _GMSTREAM_H_

#include "gmMem.h"

/// \class gmStream
/// \brief gmStream is an abstract stream class
class gmStream
{
public:

  enum
  {
    ILLEGAL_POS = -1,
  };
  
  enum Flags
  {
    F_EOS       = (1 << 0), //!< is the stream at End Of Stream
    F_READ      = (1 << 1), //!< is the stream readable
    F_WRITE     = (1 << 2), //!< is the stream writable
    F_ERROR     = (1 << 3), //!< has a stream error occured
    F_SIZE      = (1 << 4), //!< are size operations valid on this stream?
    F_SEEK      = (1 << 5), //!< are seeking operations valid on this stream?
    F_TELL      = (1 << 6), //!< are tell operations valid on this stream?
    F_USER      = (1 << 6),
  };
  
  gmStream() : m_flags(0) {}
  virtual ~gmStream() {}

  /// \brief Seek()
  /// \return the p_pos before the seek if seek is supported, else ILLEGAL_POS.
  virtual unsigned int Seek(unsigned int p_pos) = 0;

  /// \brief Tell()
  /// \return the p_pos if tell is supported, else ILLEGAL_POS
  virtual unsigned int Tell() const { return (unsigned int) ILLEGAL_POS; }

  /// \brief GetSize() will return the size of the stream if the stream supports this feature.
  /// \return ILLEGAL_POS if GetSize is not supported
  virtual unsigned int GetSize() const { return (unsigned int) ILLEGAL_POS; }

  /// \brief Read() will read p_n bytes from the stream into p_buffer.
  /// \return the number of bytes successfully read.
  virtual unsigned int Read(void * p_buffer, unsigned int p_n) = 0;

  /// \brief Write() will write p_n bytes from p_buffer to the stream.
  /// \return the number of bytes successfully written
  virtual unsigned int Write(const void * p_buffer, unsigned int p_n) = 0;

  /// \brief GetFlags() will return the current stream flags
  inline Flags GetFlags() const { return (Flags) m_flags; }

  //
  // Streaming interface (quite slow, but nice... perhaps add endian swapping templates as well)
  //
  
  template<class T>
  gmStream &operator <<(T a_v)
  {
    Write(&a_v, sizeof(T));
    return *this;
  }

  template<class T>
  gmStream &operator >>(T &a_v)
  {
    Read(&a_v, sizeof(T));
    return *this;
  }

protected:

  unsigned int m_flags;
};


#endif // _GMSTREAM_H_
